@@ -55,12 +55,16 @@ struct client {
   client() {}
   // TODO: real params
   explicit client(bool client) {
+    TPCONTEXT_T context;
+    if (tpgetctxt(&context, 0) != -1 && context >= 0) {
+      return;
+    }
+
     std::unique_ptr<char, decltype(&tpfree)> guard(
         tpalloc(const_cast<char *>("TPINIT"), nullptr, TPINITNEED(16)),
         &tpfree);
     TPINIT *tpinitbuf = reinterpret_cast<TPINIT *>(guard.get());
-    strcpy(tpinitbuf->usrname, "");
-    strcpy(tpinitbuf->passwd, "");
+    memset(tpinitbuf, 0, sizeof(*tpinitbuf));
     strcpy(tpinitbuf->cltname, "tpsysadm");
     tpinitbuf->flags = TPMULTICONTEXTS;
     if (client) {
@@ -147,6 +151,7 @@ struct xatmibuf {
   }
 
   char *p;
+
  private:
   void swap(xatmibuf &other) noexcept {
     std::swap(p, other.p);
@@ -174,14 +179,16 @@ static void default_client() {
   }
 }
 
-static py::object to_py(FBFR32 *fbfr) {
+static py::object to_py(FBFR32 *fbfr, FLDLEN32 buflen = 0) {
   FLDID32 fieldid = FIRSTFLDID;
   FLDOCC32 oc = 0;
 
   py::dict result;
   py::list val;
 
-  FLDLEN32 buflen = Fsizeof32(fbfr);
+  if (buflen == 0) {
+    buflen = Fsizeof32(fbfr);
+  }
   std::unique_ptr<char> value(new char[buflen]);
 
   for (;;) {
@@ -201,9 +208,7 @@ static py::object to_py(FBFR32 *fbfr) {
       if (name != nullptr) {
         result[name] = val;
       } else {
-        char tmpname[32];
-        snprintf(tmpname, sizeof(tmpname), "((FLDID32)%u)", fieldid);
-        result[tmpname] = val;
+        result[py::int_(fieldid)] = val;
       }
     }
 
@@ -224,13 +229,14 @@ static py::object to_py(FBFR32 *fbfr) {
         val.append(py::cast(*reinterpret_cast<double *>(value.get())));
         break;
       case FLD_STRING:
-        val.append(py::str(PyUnicode_DecodeLocale(value.get(), "surrogateescape")));
+        val.append(
+            py::str(PyUnicode_DecodeLocale(value.get(), "surrogateescape")));
         break;
       case FLD_CARRAY:
         val.append(py::bytes(value.get(), len));
         break;
       case FLD_FML32:
-        val.append(to_py(reinterpret_cast<FBFR32 *>(value.get())));
+        val.append(to_py(reinterpret_cast<FBFR32 *>(value.get()), buflen));
         break;
       default:
         throw std::invalid_argument("Unsupported field " +
@@ -264,11 +270,12 @@ static void from_py1(xatmibuf &buf, FLDID32 fieldid, FLDOCC32 oc,
     std::string val(PyBytes_AsString(obj.ptr()), PyBytes_Size(obj.ptr()));
 
     buf.mutate([&](FBFR32 *fbfr) {
-      return CFchg32(fbfr, fieldid, oc, const_cast<char *>(val.data()), val.size(),
-                     FLD_CARRAY);
+      return CFchg32(fbfr, fieldid, oc, const_cast<char *>(val.data()),
+                     val.size(), FLD_CARRAY);
     });
   } else if (py::isinstance<py::str>(obj)) {
-    py::bytes b = py::reinterpret_steal<py::bytes>(PyUnicode_EncodeLocale(obj.ptr(), "surrogateescape"));
+    py::bytes b = py::reinterpret_steal<py::bytes>(
+        PyUnicode_EncodeLocale(obj.ptr(), "surrogateescape"));
     std::string val(PyBytes_AsString(b.ptr()), PyBytes_Size(b.ptr()));
 
     buf.mutate([&](FBFR32 *fbfr) {
@@ -303,8 +310,13 @@ static void from_py(py::dict obj, xatmibuf &b) {
   xatmibuf f;
 
   for (auto it : obj) {
-    FLDID32 fieldid =
-        Fldid32(const_cast<char *>(std::string(py::str(it.first)).c_str()));
+    FLDID32 fieldid;
+    if (py::isinstance<py::int_>(it.first)) {
+      fieldid = it.first.cast<py::int_>();
+    } else {
+      fieldid =
+          Fldid32(const_cast<char *>(std::string(py::str(it.first)).c_str()));
+    }
 
     py::handle o = it.second;
     if (py::isinstance<py::list>(o)) {
@@ -364,7 +376,7 @@ static py::object pytpexport(py::object idata, long flags) {
   long olen = ostr.capacity();
   int rc = tpexport(in.p, in.len, &ostr[0], &olen, flags);
   if (rc == -1) {
-      throw xatmi_exception(tperrno);
+    throw xatmi_exception(tperrno);
   }
 
   if (flags == 0) {
@@ -377,9 +389,10 @@ static py::object pytpimport(const std::string istr, long flags) {
   xatmibuf obuf("FML32", istr.size());
 
   long olen = 0;
-  int rc = tpimport(const_cast<char*>(istr.c_str()), istr.size(), obuf.pp, &olen, flags);
+  int rc = tpimport(const_cast<char *>(istr.c_str()), istr.size(), obuf.pp,
+                    &olen, flags);
   if (rc == -1) {
-      throw xatmi_exception(tperrno);
+    throw xatmi_exception(tperrno);
   }
 
   return to_py(std::move(obuf));
@@ -503,7 +516,14 @@ void PY(TPSVCINFO *svcinfo) {
   try {
     py::gil_scoped_acquire acquire;
     auto idata = to_py(xatmibuf(svcinfo));
-    server.attr(svcinfo->name)(idata);
+
+    auto &&func = server.attr(svcinfo->name);
+    auto &&signature = py::module::import("inspect").attr("signature");
+    if (py::len(signature(func).attr("parameters")) == 3) {
+      func(idata, svcinfo->flags);
+    } else {
+      func(idata);
+    }
     if (tsvcresult.clean) {
       userlog(const_cast<char *>("tpreturn() not called"));
       tpreturn(TPEXIT, 0, nullptr, 0, 0);
@@ -605,8 +625,6 @@ static void pyrun(py::object svr, std::vector<std::string> args,
 }
 
 PYBIND11_MODULE(tuxedo, m) {
-  m.doc() = "Python3 bindings for writing Oracle Tuxedo clients and servers";
-
   // Poor man's namedtuple
   py::class_<pytpreply>(m, "TpReply")
       .def_readonly("rval", &pytpreply::rval)
@@ -708,43 +726,58 @@ PYBIND11_MODULE(tuxedo, m) {
       py::arg("message"));
 
 #if defined(TPSINGLETON) && defined(TPSECONDARYRQ)
-  m.def("tpadvertisex", &pytpadvertisex, py::arg("svcname"), py::arg("func"),
-        py::arg("flags") = 0);
+  m.def("tpadvertisex", &pytpadvertisex,
+        "Routine for advertising a service with unique service name in a "
+        "domain, or advertising a service on the secondary request queue of a "
+        "Tuxedo server.",
+        py::arg("svcname"), py::arg("func"), py::arg("flags") = 0);
 #endif
   m.def(
       "tpadvertise",
       [](const char *svcname) { pytpadvertisex(svcname, svcname, 0); },
-      py::arg("svcname"));
+      "Routine for advertising a service name", py::arg("svcname"));
 
-  m.def("run", &pyrun, py::arg("server"), py::arg("args"),
+  m.def("run", &pyrun, "Run Tuxedo server", py::arg("server"), py::arg("args"),
         py::arg("rmname") = "NONE");
 
-  m.def("tpcall", &pytpcall, py::arg("svc"), py::arg("idata"),
+  m.def("tpcall", &pytpcall,
+        "Routine for sending service request and awaiting its reply",
+        py::arg("svc"), py::arg("idata"), py::arg("flags") = 0);
+
+  m.def("tpadmcall", &pytpadmcall, "Administers unbooted application",
+        py::arg("idata"), py::arg("flags") = 0);
+
+  m.def("tpacall", &pytpacall, "Routine for sending a service request",
+        py::arg("svc"), py::arg("idata"), py::arg("flags") = 0);
+  m.def("tpgetrply", &pytpgetrply,
+        "Routine for getting a reply from a previous request", py::arg("cd"),
         py::arg("flags") = 0);
 
-  m.def("tpadmcall", &pytpadmcall, py::arg("idata"), py::arg("flags") = 0);
-
-  m.def("tpacall", &pytpacall, py::arg("svc"), py::arg("idata"),
+  m.def("tpreturn", &pytpreturn, "Routine for returning from a service routine",
+        py::arg("rval"), py::arg("rcode"), py::arg("data"),
         py::arg("flags") = 0);
-  m.def("tpgetrply", &pytpgetrply, py::arg("cd"), py::arg("flags") = 0);
+  m.def("tpforward", &pytpforward,
+        "Routine for forwarding a service request to another service routine",
+        py::arg("svc"), py::arg("data"), py::arg("flags") = 0);
 
-  m.def("tpreturn", &pytpreturn, py::arg("rval"), py::arg("rcode"),
-        py::arg("data"), py::arg("flags") = 0);
-  m.def("tpforward", &pytpforward, py::arg("svc"), py::arg("data"),
-        py::arg("flags") = 0);
-
-  m.def("tpexport", &pytpexport, py::arg("ibuf"), py::arg("flags") = 0);
-  m.def("tpimport", &pytpimport, py::arg("istr"), py::arg("flags") = 0);
+  m.def("tpexport", &pytpexport,
+        "Converts a typed message buffer into an exportable, "
+        "machine-independent string representation, that includes digital "
+        "signatures and encryption seals",
+        py::arg("ibuf"), py::arg("flags") = 0);
+  m.def("tpimport", &pytpimport,
+        "Converts an exported representation back into a typed message buffer",
+        py::arg("istr"), py::arg("flags") = 0);
 
   m.def(
       "Fldtype32", [](FLDID32 fieldid) { return Fldtype32(fieldid); },
-      py::arg("fieldid"));
+      "Maps field identifier to field type", py::arg("fieldid"));
   m.def(
       "Fldno32", [](FLDID32 fieldid) { return Fldno32(fieldid); },
-      py::arg("fieldid"));
+      "Maps field identifier to field number", py::arg("fieldid"));
   m.def(
       "Fmkfldid32", [](int type, FLDID32 num) { return Fmkfldid32(type, num); },
-      py::arg("type"), py::arg("num"));
+      "Makes a field identifier", py::arg("type"), py::arg("num"));
 
   m.def(
       "Fname32",
@@ -755,7 +788,7 @@ PYBIND11_MODULE(tuxedo, m) {
         }
         return name;
       },
-      py::arg("fieldid"));
+      "Maps field identifier to field name", py::arg("fieldid"));
   m.def(
       "Fldid32",
       [](const char *name) {
@@ -765,7 +798,71 @@ PYBIND11_MODULE(tuxedo, m) {
         }
         return id;
       },
-      py::arg("name"));
+      "Maps field name to field identifier", py::arg("name"));
+
+  m.def(
+      "Fboolev32",
+      [](py::object fbfr, const char *expression) {
+        std::unique_ptr<char, decltype(&free)> guard(
+            Fboolco32(const_cast<char *>(expression)), &free);
+        if (guard.get() == nullptr) {
+          throw fml32_exception(Ferror32);
+        }
+        auto buf = from_py(fbfr);
+        auto rc = Fboolev32(*buf.fbfr(), guard.get());
+        if (rc == -1) {
+          throw fml32_exception(Ferror32);
+        }
+        return rc == 1;
+      },
+      "Evaluates buffer against expression", py::arg("fbfr"),
+      py::arg("expression"));
+
+  m.def(
+      "Ffloatev32",
+      [](py::object fbfr, const char *expression) {
+        std::unique_ptr<char, decltype(&free)> guard(
+            Fboolco32(const_cast<char *>(expression)), &free);
+        if (guard.get() == nullptr) {
+          throw fml32_exception(Ferror32);
+        }
+        auto buf = from_py(fbfr);
+        auto rc = Ffloatev32(*buf.fbfr(), guard.get());
+        if (rc == -1) {
+          throw fml32_exception(Ferror32);
+        }
+        return rc;
+      },
+      "Returns value of expression as a double", py::arg("fbfr"),
+      py::arg("expression"));
+
+  m.def(
+      "Ffprint32",
+      [](py::object fbfr, py::object iop) {
+        auto buf = from_py(fbfr);
+        int fd = iop.attr("fileno")().cast<py::int_>();
+        std::unique_ptr<FILE, decltype(&fclose)> fiop(fdopen(dup(fd), "w"),
+                                                      &fclose);
+        auto rc = Ffprint32(*buf.fbfr(), fiop.get());
+        if (rc == -1) {
+          throw fml32_exception(Ferror32);
+        }
+      },
+      "Prints fielded buffer to specified stream", py::arg("fbfr"),
+      py::arg("iop"));
+
+  m.def(
+      "Fextread32",
+      [](py::object iop) {
+        xatmibuf obuf("FML32", 1024);
+        int fd = iop.attr("fileno")().cast<py::int_>();
+        std::unique_ptr<FILE, decltype(&fclose)> fiop(fdopen(dup(fd), "r"),
+                                                      &fclose);
+
+        obuf.mutate([&](FBFR32 *fbfr) { return Fextread32(fbfr, fiop.get()); });
+        return to_py(std::move(obuf));
+      },
+      "Builds fielded buffer from printed format", py::arg("iop"));
 
   m.attr("TPNOFLAGS") = py::int_(TPNOFLAGS);
 
@@ -838,4 +935,30 @@ PYBIND11_MODULE(tuxedo, m) {
   m.attr("BADFLDID") = py::int_(BADFLDID);
 
   m.attr("TPEX_STRING") = py::int_(TPEX_STRING);
+
+  m.doc() =
+      R"(Python3 bindings for writing Oracle Tuxedo clients and servers
+
+Flags to service routines:
+
+- TPNOBLOCK - non-blocking send/rcv
+- TPSIGRSTRT - restart rcv on interrupt
+- TPNOREPLY - no reply expected
+- TPNOTRAN - not sent in transaction mode
+- TPTRAN - sent in transaction mode
+- TPNOTIME - no timeout
+- TPABSOLUTE - absolute value on tmsetprio 
+- TPGETANY - get any valid reply
+- TPNOCHANGE - force incoming buffer to match
+- RESERVED_BIT1 - reserved for future use
+- TPCONV - conversational service
+- TPSENDONLY - send-only mode
+- TPRECVONLY - recv-only mode
+
+Flags to tpreturn:
+
+- TPFAIL - service FAILURE for tpreturn
+- TPEXIT - service FAILURE with server exit
+- TPSUCCESS - service SUCCESS for tpreturn
+)";
 }
